//---------------------------------------------------------
//          Tri trs rapide pour gros fichiers
//      - GRATUIT - (C) Pascal Peyremorte, 1996-1998
//
//                  Pour me contacter :
//
//    Pascal Peyremorte
//    Alle des Alouettes   Tel : +33+ (0)4 76 98 36 82
//    38800 Champagnier     p.peyremorte@dial.oleane.com
//         France
//
//  Tout message (questions, conseils, astuce, insultes :-)
//  sera le bienvenu !   Mise en composant Delphi 3 prvue...
//
//---------------------------------------------------------

{Mode d'emploi : Remplir les champs avec l'inspecteur ou par programme,
 puis appeler la mathode execute pour lancer le tri...}


unit FileSort;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TKeyType  =(ktSString, ktCharArray, ktByte, ktWord, ktInteger, ktUser);
  TProcComp =Function(Const X,Y):Boolean;
  TSortOptions = Set of (soReverse, soDeleteSource, soOverrideDest, soUseBufExt,
                         soBuildIndex, soIndexOnly, soKeepIndex);

  TFileSort = class(TComponent)
  private   { Dclarations prives }
    fFileIn   : AnsiString;  {Chemin et nom du ficher  trier}
    fFileOut  : AnsiString;  {Chemin et nom de fichier  crer}
    fTempPath : AnsiString;  {Chemin pour fichiers temporaires}
    fRecSize  : Integer;     {Taille de l'enregistrement}
    fKeyOffset: Integer;     {Dbut de la cl dans l'enregistrement}
    fKeySize  : Integer;     {Longueur de la cl}
    fKeyType  : TKeyType;    {Type de cl}
    fBufSize  : Integer;     {Taille des buffers}
    fOptions  : TSortOptions;{Options pour le tri}

    fMaxBuf   : Integer;     {Numro dernier enreg dans buffers}
    fBuf0     : PChar;       {Buffers de travail}
    fBuf1     : PChar;
    fBufOut   : PChar;
    fNomTmp   : AnsiString;  {Chemin+Nom des fichiers temporaires}
    fProcComp : TProcComp;   {Fonction utilise pour faire les comparaisons}

    fOnStart  : TNotifyEvent;
    fOnEnd    : TNotifyEvent;

    Procedure SetRecSize(Size:Integer);
    Procedure AjusteTailleBuffers;     {Calcule et vrifie fMaxBuf}
    Procedure SetKeyType(KT:TKeyType);
    Procedure SetOptions(Opt:TSortOptions);
  protected { Dclarations protges }
  public    { Dclarations publiques }
    Constructor Create(AOwner:TComponent); Override;
    Destructor Destroy; Override;
    Procedure SetBuffers(Buf1,Buf2,Buf3:Pointer; BufSize:Integer);
    Function Execute:Boolean;          {Lance le tri}
    Procedure SetUserCompare(FComp:TProcComp);
  published { Dclarations publies }
    Property FileIn   :AnsiString read fFileIn write fFileIn;
    Property FileOut  :AnsiString read fFileOut write fFileOut;
    Property TempPath :AnsiString read fTempPath write fTempPath;
    Property RecSize  :Integer read fRecSize write SetRecSize;
    Property KeyOffset:Integer read fKeyOffset write fKeyOffset;
    Property KeySize  :Integer read fKeySize write fKeySize;
    Property KeyType  :TKeyType read fKeyType write SetKeyType;
    Property BufSize  :Integer read fBufSize write fBufSize default 65536;
    Property Options  :TSortOptions read fOptions write SetOptions;

    Property OnStartSort  :TNotifyEvent read fOnStart write fOnStart;
    Property OnEndSort    :TNotifyEvent read fOnEnd   write fOnEnd;
  end;

  EFileSort = Class(Exception);

procedure Register;

{-------------------------------------------------------------------------}

implementation

procedure Register;
begin
  RegisterComponents('Perso', [TFileSort]);
end;

//*********************************************************************
//   Fonctions de comparaisons
//*********************************************************************

Function CompareByteAZ(Const X,Y):Boolean; Register;
Begin
  Result:= Byte(X)>Byte(Y);
End;
Function CompareByteZA(Const X,Y):Boolean; Register;
Begin
  Result:= Byte(X)<Byte(Y);
End;

Function CompareWordAZ(Const X,Y):Boolean; Register;
Begin
  Result:= Word(X)>Word(Y);
End;
Function CompareWordZA(Const X,Y):Boolean; Register;
Begin
  Result:= Word(X)<Word(Y);
End;

Function CompareIntegerAZ(Const X,Y):Boolean; Register;
Begin
  Result:= Integer(X)>Integer(Y);
End;
Function CompareIntegerZA(Const X,Y):Boolean; Register;
Begin
  Result:= Integer(X)<Integer(Y);
End;

Function CompareSStringAZ(Const X,Y):Boolean; Register;
Begin
  Result:= ShortString(X)>ShortString(Y);
End;
Function CompareSStringZA(Const X,Y):Boolean; Register;
Begin
  Result:= ShortString(X)<ShortString(Y);
End;

Function CompareArrayAZ(Const X,Y):Boolean; Register;
Begin
  Result:= ShortString(X)>ShortString(Y);
End;
Function CompareArrayZA(Const X,Y):Boolean; Register;
Begin
  Result:= ShortString(X)<ShortString(Y);
End;

Const TabProcComp : Array[ktSSTring..ktInteger,Boolean] of TProcComp =
        ((CompareSStringAZ,CompareSStringZA),
         (CompareArrayAZ,CompareArrayZA),
         (CompareByteAZ, CompareByteZA),
         (CompareWordAZ,CompareWordZA),
         (CompareIntegerAZ,CompareIntegerZA));

//*********************************************************************
//   TFileSort : Construction / destruction
//*********************************************************************

Constructor TFileSort.Create(AOwner:TComponent);
Var
   Chemin  : PChar;
Begin
  Inherited Create(AOwner);

  {Rcupre le rpertoire des fichiers temporaire}
  GetMem(Chemin,MAX_PATH);
  GetTempPath(MAX_PATH,Chemin);
  fTempPath:=AnsiString(Chemin);
  If fTempPath[Length(fTempPath)]<>'\' then fTempPath:=fTempPath+'\';
  FreeMem(Chemin);

  fBufSize:=65536;
  fOptions:=[];
End;

Destructor TFileSort.Destroy;
Begin
  Inherited Destroy;
End;

//*********************************************************************
//   TFileSort : Gestion des paramtres
//*********************************************************************

Procedure TFileSort.AjusteTailleBuffers;
Begin
  If fRecSize<>0 then
    fMaxBuf:=(fBufSize div fRecSize)* fRecSize;
End;

Procedure TFileSort.SetRecSize(Size:Integer);
Begin
  fRecSize:=Size;
  AjusteTailleBuffers;
End;

Procedure TFileSort.SetBuffers(Buf1,Buf2,Buf3:Pointer; BufSize:Integer);
Begin
  Include(fOptions, soUseBufExt);  {Indique buffers fournis par utilisateurs}
  fBuf0:=Buf1;        {Garde copie des pointeurs}
  fBuf1:=Buf2;
  fBufOut:=Buf3;
  fBufSize:=BufSize;  {et de la taille des buffers fournis}
  AjusteTailleBuffers;{Calcule nombre d'enregs. par buffer}
End;

Procedure TFileSort.SetKeyType(KT:TKeyType);
Begin
  fKeyType:=KT;
  Case KT of
    ktByte: {la cl est un octet}
      fKeySize:=1;
    ktWord: {La cl est un word}
      fKeySize:=2;
    ktInteger: {La cl est un integer}
      fKeySize:=4;
  End;
End;

Procedure TFileSort.SetOptions(Opt:TSortOptions);
Begin
  fOptions:=Opt;
End;

Procedure TFileSort.SetUserCompare(FComp:TProcComp);
Begin
  fProcComp:=FComp;
End;

//*********************************************************************
//   TFileSort : Procdure principale
//*********************************************************************

Function TFileSort.Execute:Boolean; {Lance le tri}
TYPE
   TNomF = String[12];
VAR
   TempName : PChar;
   FichIn   :File;      {Fichier d'entre}

   F0,F1,               {Fichiers temporaires tri et fusion}
   FOut  : File;        {Fichier de sortie fusion}

   Niveau: Word;
   Fini  : Boolean;

   Ref   : Pointer;     {Variables de TriMem non rcusives mise ici pour}
   Swap  : Pointer;     {ne pas encombrer la pile lors des rentrances}

   {-----------------}

   {Tri du buffer fBuf0 en mmoire par quicksort. Sans surprises}
   {Attention : I ou J peuvent tre ngatifs}
   Procedure TriMem(D,F:Pointer);
   Var
     I,J : PChar;
   Begin
     {Prendre un enregistrement au milieu, Attention si nombre impair}
     If (((Integer(F)-Integer(D)) div fRecSize) and 1)=0
       then I:=Pointer((Integer(D)+Integer(F)) div 2)
       else I:=Pointer((Integer(D)+Integer(F)+fRecSize) div 2);
     {Note : D+F ne dborde jamais car la zone mmoire au del de $80000000
      est rserve aux fichiers mapps et au syst. d'expoitation}

     Move(I[fKeyOffset],Ref^,fKeySize);
     I:=D; J:=F;
     Repeat
       Inc(I,fKeyOffset); Inc(J,fKeyOffset);
       While fProcComp(Ref^,I^) DO Inc(I,fRecSize);
       While fProcComp(J^,Ref^) DO Dec(J,fRecSize);
       Dec(I,fKeyOffset); Dec(J,fKeyOffset);
       If I<=J then
       Begin
         If I<>J then
         Begin
           move(I^,   Swap^, fRecSize);
           move(J^,   I^,    fRecSize);
           move(Swap^,J^,    fRecSize);
         End;
         Inc(I,fRecSize);
         Dec(J,fRecSize);
       End;
     Until I>J;

     If J-D<F-I then                  {Commence par le tri le plus court}
     Begin                            {pour rduire l'emploi de la pile }
       If D<J then TriMem(D,J);
       If I<F then TriMem(I,F);
     End
     else
     Begin
       If I<F then TriMem(I,F);
       If D<J then TriMem(D,J);
     End
   End;

   {-----------------}

   Function CarHex(I:Byte):Char; Assembler; Register;
   Asm
     add  al,'0'
     cmp  al,'9'
     jbe  @Fin
     add  al,7
   @Fin:
   End;

   Function NomFicTemp(N,x:Byte):AnsiString;
   Begin
     Result:=ChangeFileExt(fNomTmp,'.'+CarHex(Hi(N))+CarHex(Lo(N))+Char(48+x));
   End;

   {-----------------}

   {Fusion de deux fichiers en un seul, en respectant le tri}
   Procedure Fusion(N,x:Byte);
   Var Lus0,Lus1 : Integer;
       Fini0,Fini1 : Boolean;

       Ptr0,Ptr1, PtrOut,
       Fin0,Fin1, FinOut : PChar;

       Procedure ReadBuf0;
       Begin
         BlockRead(F0,fBuf0^,fMaxBuf,Lus0);
         Fini0:= Lus0=0;
         Ptr0:=fBuf0;
         Fin0:=@fBuf0[Lus0];
       End;

       Procedure ReadBuf1;
       Begin
         BlockRead(F1,fBuf1^,fMaxBuf,Lus1);
         Fini1:= Lus1=0;
         Ptr1:=fBuf1;
         Fin1:=@fBuf1[Lus1];
       End;

       Procedure WriteBufOut;
       Begin
         BlockWrite(FOut,fBufOut^,Integer(PtrOut-fBufOut));
         PtrOut:=fBufOut;
       End;
   Begin
     AssignFile(F0,NomFicTemp(N-1,0));
     AssignFile(F1,NomFicTemp(N-1,1));
     Reset(F0,1);
     Reset(F1,1);

     AssignFile(FOut,NomFicTemp(N,x));
     Rewrite(FOut,1);

     ReadBuf0;
     ReadBuf1;
     PtrOut:=fBufOut;
     FinOut:=@fBufOut[fMaxBuf];
     Repeat
       While Not Fini0 and
            (Fini1 or not fProcComp(Ptr0[fKeyOffset],Ptr1[fKeyOffset])) do
       Begin
         Move(Ptr0^,PtrOut^,fRecSize);
         Inc(Ptr0,fRecSize);
         If Ptr0>=Fin0 then ReadBuf0;
         Inc(PtrOut,fRecSize);
         If PtrOut>=FinOut then WriteBufOut;
       End;

       While Not Fini1 and
            (Fini0 or not fProcComp(Ptr1[fKeyOffset],Ptr0[fKeyOffset])) do
       Begin
         Move(Ptr1^,PtrOut^,fRecSize);
         Inc(Ptr1,fRecSize);
         If Ptr1>=Fin1 then ReadBuf1;
         Inc(PtrOut,fRecSize);
         If PtrOut>=FinOut then WriteBufOut;
       End;
     Until Fini0 and Fini1;
     If PtrOut>fBufOut then WriteBufOut;

     CloseFile(FOut);
     CloseFile(F1);
     Erase(F1);
     CloseFile(F0);
     Erase(F0);
   End;

   {-----------------}

   {Tri et fusions rcursives}
   Procedure Tri(N,x:Byte);
   Var Lus:Integer;
   Begin
     If N=0 then
     Begin
       BlockRead(FichIn,fBuf0^,fMaxBuf,Lus);
       Fini:=Lus<>fMaxBuf;
       TriMem(@fBuf0^,@fBuf0[Lus-fRecSize]); {Tri les N enreg numrots de 0  N-1}

       AssignFile(Fout,NomFicTemp(N,x));
       Rewrite(Fout,1);
       BlockWrite(Fout,fBuf0^,Lus);
       CloseFile(Fout);
     End
     else
     Begin
       Tri(N-1,0);
       If Fini then
         Rename(FOut,NomFicTemp(N,x))
       else
       Begin
         Tri(N-1,1);
         Fusion(N,x); {Fusion de (N-1,0) et (N-1,1) en (N,x)}
       End;
     End;
   End;

   {-----------------}

Begin
  Result:=False;
  {Verifie fichier d'entre prsent}
  If Not FileExists(fFileIn) then
    Raise EFileSort.Create('Source file absent');
  If fRecSize=0 then
    Raise EFileSort.Create('RecSize is nul');
  If fKeySize=0 then
    Raise EFileSort.Create('KeySize is nul');
  If (fKeyOffset<0) or (fKeyOffset+fKeySize>fRecSize) then
    Raise EFileSort.Create('Key out of Record');

  Try
    {Alloue les buffers s'il ne sont pas fournis par l'utilisateur}
    If not (soUseBufExt in fOptions) then
    Begin
      GetMem(fBuf0,fBufSize);  {??? Grer erreur allocation}
      GetMem(fBuf1,fBufSize);
      GetMem(fBufOut,fBufSize);
    End;

    {Alloue les variables de swap et de comparaison}
    GetMem(Ref,fKeySize);
    GetMem(Swap,fRecSize);      {??? Grer erreur allocation}

    {Cherche un nom pour les fichiers temporaires}
    GetMem(TempName,MAX_PATH);
    GetTempFileName(PChar(fTempPath),'PPS',0,TempName);
    fNomTmp:=TempName;
    FreeMem(TempName);

    {Met en place la procdure de comparaison sauf si mise par utilisateur}
    If not assigned(fProcComp) then
      fProcComp:=TabProcComp[fKeyType, (soReverse in fOptions)];

    {Ouvre fichier et dbute le tri}
    AssignFile(FichIn,fFileIn);
    Reset(FichIn,1);          {Ouvre le fichier  trier}
    If FileSize(FichIn) div fRecSize <2 then
      Raise EFileSort.Create('Need at least 2 records');

    If Assigned(fOnStart) then fOnStart(Self);

    Fini:=False;              {Raz indicateur tri fini}
    Niveau:=0;                {Commence par rcusion n0}
    Tri(Niveau,0);            {Gnre niveau N, premire moiti}
    While not Fini do
    Begin
      Tri(Niveau,1);          {Gnre niveau N, seconde moiti}
      Inc(Niveau);            {Gnre niveau N+1, premire moiti... }
      Fusion(Niveau,0)        {... par fusion des 2 moitis de niveau N}
    End;

    CloseFile(FichIn);        {Ferme le fichier initial}

    {Supprime l'original si option choisie}
    If soDeleteSource in fOptions then
      Erase(FichIn);

    {Vrifie fichier de sortie absent}
    If FileExists(FileOut) then
      If soOverrideDest in fOptions
        then DeleteFile(FileOut)
        else Raise EFileSort.Create('Dest. file already present');

    {Renomme/dplace le fichier de sortie au nom de l'initial}
    MoveFile(TFileRec(FOut).Name,PChar(fFileOut));

    {Si procdure de sortie, l'excute}
    If Assigned(fOnEnd) then fOnEnd(Self);

    Result:=True; {tout c'est bien pass}
  Finally
    {Libre tous les buffers et variables}
    FreeMem(Ref);
    FreeMem(Swap);
    If Not (soUseBufExt in fOptions) then
    Begin
      FreeMem(fBuf0,fBufSize);
      FreeMem(fBuf1,fBufSize);
      FreeMem(fBufOut,fBufSize);
    End;
  End;
End;

end.
