unit U_TestFS;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ComCtrls, ExtCtrls,
  FileSort;

type
  TEcrPrin = class(TForm)
    FileSort1: TFileSort;
    ListIn: TListBox;
    ListOut: TListBox;
    BtnScan: TBitBtn;
    BtnTrie: TBitBtn;
    Disques: TEdit;
    Label1: TLabel;
    TypeTri: TRadioGroup;
    Label2: TLabel;
    Label3: TLabel;
    lbKS: TLabel;
    lbKT: TLabel;
    edKeyType: TComboBox;
    edKeySize: TEdit;
    edKeyOffset: TEdit;
    lbKO: TLabel;
    GroupBox1: TGroupBox;
    cbReverse: TCheckBox;
    cbDelSource: TCheckBox;
    cbOverrideDest: TCheckBox;
    cbUseBufExt: TCheckBox;
    Resultat: TListBox;
    procedure BtnScanClick(Sender: TObject);
    procedure BtnTrieClick(Sender: TObject);
    procedure TypeTriClick(Sender: TObject);
    procedure cbReverseClick(Sender: TObject);
    procedure cbDelSourceClick(Sender: TObject);
    procedure cbOverrideDestClick(Sender: TObject);
    procedure cbUseBufExtClick(Sender: TObject);
    procedure edKeyTypeChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FileSort1EndSort(Sender: TObject);
    procedure FileSort1StartSort(Sender: TObject);
  private { Dclarations prives }
    Procedure MetInfo(Const T:String);
    Procedure Affiche(Fichier:String; Liste:TListBox; Info:String);
    Procedure AfficheParams;
    Procedure UdateOptions(Opt:TSortOptions; Usage:Boolean);
  public  { Dclarations publiques }
  end;

  TNom    = String[255];        {Place pour nom de fichier}
  TRef    = Record
              Nom    :TNom;      {Nom du fichier}
              Taille :LongInt;   {Taille du fichier}
              Rep    :Word;      {Num rp. associ au fichier}
            End;
Const
   OffsetNom   =0;
   OffsetTaille=SizeOf(TNom);
   OffsetRep   =SizeOf(LongInt) + OffsetTaille;

var
  EcrPrin: TEcrPrin;

implementation

Var
   NbrFic  : Integer;        {Nombre de fichiers trouv}

{$R *.DFM}

//*********************************************************************
//   Divers affichage pour visualisrer options
//*********************************************************************

Procedure TEcrPrin.MetInfo(Const T:String);
Begin
  Resultat.ItemIndex:=Resultat.Items.Add(T);
End;

Procedure TEcrPrin.AfficheParams;
Begin
  edKeyType.ItemIndex:=Byte(FileSort1.KeyType);
  edKeySize.Text:=IntToStr(FileSort1.KeySize);
  edKeyOffset.Text:=IntToStr(FileSort1.KeyOffset);
  cbReverse.Checked:=soReverse in FileSort1.Options;
  cbDelSource.Checked:=soDeleteSource in FileSort1.Options;
  cbOverrideDest.Checked:=soOverrideDest in FileSort1.Options;
  cbUseBufExt.Checked:=soUseBufExt in FileSort1.Options;
End;

Procedure TEcrPrin.Affiche(Fichier:String; Liste:TListBox; Info:String);
Var
   F:File of TRef;
   Buf:TRef;
Begin
  MetInfo(Info);
  Application.ProcessMessages;
  Screen.Cursor:=crHourGlass;
  Try
    AssignFile(F,Fichier);
    Reset(F);
    Liste.Clear;
    While Not EOF(F) do
    Begin
      Read(F,Buf);
      Liste.Items.Add(Format('%8x  %4x  %s',[Buf.Taille,Buf.Rep,Buf.Nom]));
    End;
    CloseFile(F);
  Finally
    Screen.Cursor:=crDefault;
  End;
End;


//*********************************************************************
//   Cration destruction de la fiche
//*********************************************************************

procedure TEcrPrin.FormShow(Sender: TObject);
begin
  AfficheParams;
end;



//*********************************************************************
//   Collecte tous les noms de fichiers des disques prsents
//*********************************************************************

procedure TEcrPrin.BtnScanClick(Sender: TObject);
Var
   I       : Integer;
   Liste   : String;         {Liste des reps  scanner}
   Chemin  : String;         {Chemin  scanner}
   NbrFic  : Integer;        {Nombre de fichiers trouv}
   NbrRep  : Integer;        {Nombre de rpertoires trouvs}

   F       : File of TRef;   {Fichier qui sera  trier}

   {Variable locale de TraiteChemin, mise ici pour ne pas surcharger pile}
   Buf     : TRef;

   Procedure TraiteChemin(Repert:String);
   Var
      NumRep : Word;
      S      : TSearchRec;
   Begin
     Caption:=Repert;
     Repert:=Repert+'\';
     NumRep:=NbrRep;         {garde n du chemin}
     Inc(NbrRep);            {et compte le chemin sauv}

     If FindFirst(Repert+'*.*',faAnyFile,S)=0 then
     Repeat
       If (S.Name<>'.') and (S.Name<>'..') then
       Begin
         If Boolean(S.Attr and faDirectory) then
           TraiteChemin(Repert+S.Name)
         else
         If not Boolean(S.Attr and faVolumeID) then
         Begin
           Buf.Nom:=S.Name;
           Buf.Taille:=S.Size;
           Buf.Rep:=NumRep;
           Write(F,Buf);
           Inc(NbrFic);
         End;
       End;
     until FindNext(S)<>0;
     FindClose(S);
   End;
Begin
  AssignFile(F, FileSort1.FileIn);
  Rewrite(F);       {Cre fichier des noms et rfrences}
  NbrFic:=0;
  NbrRep:=0;        {... ni de rpertoires}

  Liste:=Disques.Text;
  While Liste<>'' do
  Begin
    I:=Pos(';',Liste);      {Extrait un chemin de la liste}
    If I=0 then I:=Length(Liste);
    Chemin:=Trim(Copy(Liste,1,I-1));
    If Chemin[Length(Chemin)]='\' then
      SetLength(Chemin,Length(Chemin)-1);
    Delete(Liste,1,I);

    TraiteChemin(Chemin);   {Scanne le chemin en quaestion}
  End;
  CloseFile(F);
  MetInfo(IntToStr(NbrFic)+' fichiers dans '+IntToStr(NbrRep)+' rpertoires');
  Affiche(FileSort1.FileIn, ListIn,
          'Chargement du fichier source dans la TListBox...');
End;


//*********************************************************************
//   Saisie des options
//*********************************************************************

procedure TEcrPrin.TypeTriClick(Sender: TObject);
begin
  Case TypeTri.ItemIndex of
    0:Begin  {Tri par noms}
        FileSort1.KeyType:=ktSString;
        FileSort1.KeyOffset:=OffsetNom;
        FileSort1.KeySize:=SizeOf(TNom); {256}
      End;
    1:Begin  {Tri par taille}
        FileSort1.KeyType:=ktInteger;
        FileSort1.KeyOffset:=OffsetTaille;
     // FileSort1.KeySize:=SizeOf(Integer); dj mis par KeyType=ktInteger
      End;
    2:Begin  {Tri par rpertoire}
        FileSort1.KeyType:=ktInteger;
        FileSort1.KeyOffset:=OffsetRep;
     // FileSort1.KeySize:=SizeOf(Integer); dj mis par KeyType=ktInteger
      End;
  End;
  AfficheParams;
end;

procedure TEcrPrin.edKeyTypeChange(Sender: TObject);
begin
  FileSort1.KeyType:=TKeyType((Sender as TComboBox).ItemIndex);
  AfficheParams;
end;

Procedure TEcrPrin.UdateOptions(Opt:TSortOptions; Usage:Boolean);
Begin
  If Usage
    then FileSort1.Options:=FileSort1.Options + Opt
    else FileSort1.Options:=FileSort1.Options - Opt;
  AfficheParams;
End;

procedure TEcrPrin.cbReverseClick(Sender: TObject);
begin
  UdateOptions([soReverse], (Sender as TCheckBox).Checked);
end;

procedure TEcrPrin.cbDelSourceClick(Sender: TObject);
begin
  UdateOptions([soDeleteSource], (Sender as TCheckBox).Checked);
end;

procedure TEcrPrin.cbOverrideDestClick(Sender: TObject);
begin
  UdateOptions([soOverrideDest], (Sender as TCheckBox).Checked);
end;

procedure TEcrPrin.cbUseBufExtClick(Sender: TObject);
begin
  UdateOptions([soUseBufExt], (Sender as TCheckBox).Checked);
end;

//*********************************************************************
//   Procedures du tri
//*********************************************************************

procedure TEcrPrin.BtnTrieClick(Sender: TObject);
Var
  S: TSearchRec;
  N: Integer;
  T: TDateTime;
begin
  FindFirst(FileSort1.FileIn, faAnyfile ,S);
  N:=S.Size;
  FindClose(S);
  NbrFic:=N div FileSort1.RecSize;

  MetInfo('Tri de '+IntToStr(NbrFic)+' enregistrements '+
          'dans '+IntToStr(FileSort1.RecSize*NbrFic)+' octets');
  T:=Time;
  FileSort1.Execute;
  MetINfo('le tri a dur '+IntToStr(Round(Frac(Time-T)*24*60*60*1000))+' ms');
  Affiche(FileSort1.FileOut, ListOut, 'Chargement de la TListBox trie');
end;

procedure TEcrPrin.FileSort1StartSort(Sender: TObject);
begin
  MetInfo('Dbut du tri  '+TimeToStr(Time));
end;

procedure TEcrPrin.FileSort1EndSort(Sender: TObject);
begin
  MetInfo('Fin du tri    '+TimeToStr(Time));
end;

end.
